from typing import List

from scenario_gym.entity import Entity, Pedestrian
from scenario_gym.pedestrian.observation import PedestrianObservation
from scenario_gym.sensor import Sensor
from scenario_gym.state import State


class PedestrianSensor(Sensor):
    """Returns observation (complete state) for pedestrian entities."""

    def __init__(
        self,
        entity: Entity,
        head_rot_angle: float = 0.0,
        distance_threshold: float = 1.0,
    ):
        """
        Init the pedestrian sensor.

        Parameters
        ----------
        entity : Entity
            The pedestrian entity.

        head_rot_angle : float
            Rotation angle of pedestrian head in rad wrt to heading
            (0 means looking forward).

        distance_threshold: float
            Only pedestrians within this distance of the entity will be considered.

        """
        super().__init__(entity)
        self.head_rot_angle = head_rot_angle
        self.distance_threshold = distance_threshold

    def _reset(self, state: State) -> PedestrianObservation:
        """Reset the sensor."""
        return self._step(state)

    def _step(self, state: State) -> PedestrianObservation:
        """Produce the pedestrian observation."""
        near_peds = self.get_nearby_pedestrians(state)
        return PedestrianObservation(
            self.entity,
            *state.get_entity_data(self.entity),
            self.head_rot_angle,
            near_peds,
            state.scenario.road_network.walkable_surface,
            state.scenario.road_network.impenetrable_surface,
        )

    def get_nearby_pedestrians(self, state: State) -> List[Entity]:
        """Get other pedestrians within a radius of the entity."""
        return [
            (e, state.poses[e], state.velocities[e])
            for e in state.get_entities_in_radius(
                *state.poses[self.entity][:2],
                self.distance_threshold,
            )
            if (isinstance(e, Pedestrian) or (e.type == "Pedestrian"))
            and (e != self.entity)
        ]



from .base import Sensor
from .common import (
    CombinedSensor,
    EgoLocalizationSensor,
    FutureCollisionDetector,
    GlobalCollisionDetector,
    KeyboardInputDetector,
)
from .map import RasterizedMapSensor



from typing import Optional

import numpy as np


class Action:
    """Base class for actions that agents commnicate to controllers."""

    pass


class TeleportAction(Action):
    """An action consiting of desired coordinates for the next pose."""

    def __init__(
        self,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 0.0,
        h: float = 0.0,
        r: float = 0.0,
        p: float = 0.0,
        pose: Optional[np.ndarray] = None,
    ):
        """
        Teleport action from coordinates or a pose.

        Parameters
        ----------
        x : float
            The x coordinate of the action.

        y : float
            The y coordinate of the action.

        z : float
            The z coordinate of the action.

        h : float
            The h coordinate of the action.

        r : float
            The r coordinate of the action.

        p : float
            The p coordinate of the action.

        pose : Optional[np.ndarray]
            The whole pose as a numpy array of shape (6,). Will
            overwrite any other coordinates passed.

        """
        self.x = pose[0] if pose is not None else x
        self.y = pose[1] if pose is not None else y
        self.z = pose[2] if pose is not None else z
        self.h = pose[3] if pose is not None else h
        self.r = pose[4] if pose is not None else r
        self.p = pose[5] if pose is not None else p

    @property
    def pose(self) -> np.ndarray:
        """Return a pose representation of the action as an array of shape (6,)."""
        return np.array([self.x, self.y, self.z, self.h, self.r, self.p])


class VehicleAction(Action):
    """An acceleration and a steering update."""

    def __init__(self, accel: float, steer: float):
        """
        Vehicle action from acceleration and steering.

        Parameters
        ----------
        accel : float
            The acceleration of the vehicle.

        steer : float
            The steering angle of angular velocity of the vehicle.

        """
        self.acceleration = accel
        self.steering = steer



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Type, TypeVar

State = TypeVar("State")


class StateCallback(ABC):
    """
    A callback to provide additional information to the state.

    Callbacks are updated during each timestep and should modify their own internal
    state in the `__call__` method.
    """

    required_callbacks: List[Type[StateCallback]] = []

    def __init__(self):
        self.callbacks: List[StateCallback] = []

    def reset(self, state: State) -> None:
        """Reset the callback check dependents are there."""
        self.callbacks.clear()
        pre_callbacks = []
        for cb in state.state_callbacks:
            if cb is self:
                break
            if cb.__class__ in self.required_callbacks:
                pre_callbacks.append(cb)

        missing = set(self.required_callbacks).difference(
            (cb.__class__ for cb in pre_callbacks)
        )
        if missing:
            raise ValueError(
                f"Cannot run callback {self.__class__.__name__} without callbacks"
                f" {missing}."
            )

        self.callbacks.extend(pre_callbacks)
        self._reset(state)

    def _reset(self, state: State) -> None:
        """Reset the callback's parameters."""
        pass

    @abstractmethod
    def __call__(self, state: State) -> None:
        """Update callback with new information."""
        raise NotImplementedError



from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from shapely.geometry import Polygon
from shapely.strtree import STRtree

try:
    from functools import cached_property
except ImportError:

    def cached_property(fn):
        """Replace cached_property with a size 1 lru cache."""
        return property(lru_cache(maxsize=1)(fn))


try:
    from numpy.typing import ArrayLike, NDArray
except ImportError:
    ArrayLike = NDArray = np.ndarray


ArgsKwargs = Tuple[Tuple[Any, ...], Dict[str, Any]]


def detect_geom_collisions(
    geoms: List[Polygon],
    others: Optional[List[Polygon]] = None,
) -> Dict[str, List[Polygon]]:
    """
    Detect collisions between polygons.

    Parameters
    ----------
    geoms : List[Polygon]
        The geometries to use.

    others : Optional[List[Polygon]]
        Additional geometries to include when checking for collisions
        with each geometry in geoms.

    Returns
    -------
    Dict[str, List[Polygon]]
        A dictionary that maps the id of each polygon in geoms to the
        polygons that it intersects.

    """
    all_geoms = geoms if others is None else geoms + others
    tree = STRtree(all_geoms)
    return {
        g: [
            g_prime
            for g_prime in tree.geometries.take(
                tree.query(g, predicate="intersects").tolist()
            )
            if g != g_prime
        ]
        for g in geoms
    }



"""Provides a selection of commonly used sensors."""
from dataclasses import dataclass
from typing import Dict, List

import numpy as np

from scenario_gym.entity import Entity
from scenario_gym.observation import (
    Observation,
    SingleEntityObservation,
    combine_observations,
)
from scenario_gym.state import State, detect_collisions

from .base import Sensor


class CombinedSensor(Sensor):
    """Combines different observations into one."""

    def __init__(self, entity: Entity, *sensors: Sensor):
        """Init the sensor."""
        assert [s.entity == entity for s in sensors]
        super().__init__(entity)
        self.sensors = sensors
        self.obs_class = None

    def _reset(self, state: State) -> Observation:
        """Reset all sensors."""
        init_obs = [s.reset(state) for s in self.sensors]
        self.obs_class = combine_observations(*(obs.__class__ for obs in init_obs))
        return self.obs_class.from_obs(*init_obs)

    def _step(self, state: State) -> Observation:
        """Get observations from all sensors."""
        return self.obs_class.from_obs(*(s.step(state) for s in self.sensors))


class EgoLocalizationSensor(Sensor):
    """Observation containing just the base entity information."""

    def _reset(self, state: State) -> SingleEntityObservation:
        """Return the entity observation."""
        return self._step(state)

    def _step(self, state: State) -> SingleEntityObservation:
        """Return the entity observation."""
        return SingleEntityObservation(
            self.entity, *state.get_entity_data(self.entity)
        )


@dataclass
class FutureCollisionObservation(SingleEntityObservation):
    """Observation with future collision information."""

    future_collision: bool


class FutureCollisionDetector(Sensor):
    """
    Detects any future collisions with the ego in the scenario.

    Entity trajectories are used to obtain their future position
    and compare t to the sensor's entity.
    """

    def __init__(self, entity: Entity, horizon: float = 5.0):
        """
        Init the sensor.

        Parameters
        ----------
        entity : Entity
            The entity.

        horizon : float
            The time horizon over which to look for collisions.

        """
        super().__init__(entity)
        self.horizon = horizon

    def _reset(self, state: State) -> FutureCollisionObservation:
        """Return future collisions."""
        return self._step(state)

    def _step(self, state: State) -> FutureCollisionObservation:
        """Return future collisions."""
        ents = {e: None for e in state.scenario.entities if e != self.entity}

        # check for collisions over the horizon
        future_collision = False
        for t in np.linspace(state.t, state.t + self.horizon, 10):
            ego_pose = self.entity.trajectory.position_at_t(t)
            for e in ents:
                ents[e] = e.trajectory.position_at_t(t)
            collisions = detect_collisions({self.entity: ego_pose}, ents)
            if len(collisions[self.entity]) > 0:
                future_collision = True
        return FutureCollisionObservation(
            self.entity,
            *state.get_entity_data(self.entity),
            future_collision,
        )


@dataclass
class CollisionObservation(SingleEntityObservation):
    """Observation with detected collisions."""

    collisions: Dict[Entity, List[Entity]]


class GlobalCollisionDetector(Sensor):
    """Returns collisions observed in the scene."""

    def _reset(self, state: State) -> CollisionObservation:
        """Return the collision observation."""
        return self._step(state)

    def _step(self, state: State) -> CollisionObservation:
        """Return the collision observation."""
        return CollisionObservation(
            self.entity,
            *state.get_entity_data(self.entity),
            state.collisions(),
        )


@dataclass
class KeyboardObservation(SingleEntityObservation):
    """Observation with detected collisions."""

    last_keystroke: Dict[Entity, List[Entity]]


class KeyboardInputDetector(Sensor):
    """Detects keyboard input."""

    def _reset(self, state: State) -> KeyboardObservation:
        """Return the collision observation."""
        return self._step(state)

    def _step(self, state: State) -> KeyboardObservation:
        """Return the keyboard observation."""
        return KeyboardObservation(
            self.entity,
            *state.get_entity_data(self.entity),
            state.last_keystroke,
        )



from typing import Dict, List, Optional, Tuple
from uuid import uuid4

import numpy as np
from pyxodr.road_objects.lane import Lane as xodrLane
from pyxodr.road_objects.network import RoadNetwork as xodrRoadNetwork
from shapely.geometry import LineString, Polygon

from scenario_gym.road_network import Lane, LaneType, Road


def xodr_lane_to_sg(
    lane: xodrLane,
    simplify_tolerance: float,
) -> Optional[Lane]:
    """Convert an OpenDRIVE lane to a scenario_gym lane."""
    if lane.type is None:
        return None

    lane_traffic_flow_line = LineString(lane.traffic_flow_line[:, :2])
    if simplify_tolerance is not None:
        lane_traffic_flow_line = lane_traffic_flow_line.simplify(simplify_tolerance)

    elevation = lane.traffic_flow_line
    boundary = np.vstack(
        [
            lane.boundary_line,
            np.flip(lane.lane_offset_line, axis=0),
        ]
    )
    lane_boundary = Polygon(boundary)
    try:
        lane_type = LaneType[lane.type]
    except KeyError as e:
        raise KeyError(
            f"Lane {lane.id} has an invalid lane type: {lane.type}. Allowed lane "
            f"types are {list(LaneType.__members__)}."
        ) from e

    return Lane(
        str(uuid4()),
        lane_boundary,
        lane_traffic_flow_line,
        [],
        [],
        lane_type,
        elevation=elevation,
    )


def road_to_sg(
    xodr_road,
    simplify_tolerance: float,
) -> List[Road]:
    """Convert an OpenDRIVE road into a list of roads."""
    roads, old_to_new_lanes = [], {}
    for _, xodr_lane_section in enumerate(xodr_road.lane_sections):
        if simplify_tolerance is not None:
            (
                x_boundary,
                y_boundary,
            ) = xodr_lane_section.boundary.exterior.simplify(simplify_tolerance).xy
        else:
            x_boundary, y_boundary = xodr_lane_section.boundary.exterior.xy

        road_boundary = Polygon(list(zip(x_boundary, y_boundary)))

        xyz_centre = xodr_lane_section.get_offset_line()
        road_center = LineString(xyz_centre[:, :2])
        if simplify_tolerance is not None:
            road_center = road_center.simplify(simplify_tolerance)
        road_elevation = xyz_centre

        lanes = []
        for lane in xodr_lane_section.lanes:
            sg_lane = xodr_lane_to_sg(lane, simplify_tolerance)
            if sg_lane is not None:
                lanes.append(sg_lane)
                old_to_new_lanes[lane] = sg_lane

        road = Road(
            str(uuid4()),
            road_boundary,
            road_center,
            lanes=lanes,
            elevation=road_elevation,
        )
        roads.append(road)

    return roads, old_to_new_lanes


def add_connection(conn: Tuple[Lane, Lane]) -> None:
    """Connect a pair of successive lanes."""
    pre, suc = conn
    if pre not in suc.predecessors:
        suc.predecessors.append(pre)
    if suc not in pre.successors:
        pre.successors.append(suc)


def xodr_to_sg_roads(
    road_network: xodrRoadNetwork,
    simplify_tolerance: float,
) -> List[Road]:
    """
    Convert a pyxodr road network to a list of roads.

    Parameters
    ----------
    road_network : xodrRoadNetwork
        Imported OpenDRIVE file.

    simplify_tolerance : float
        Points per m for simplifying center and boundary lines.

    """
    xodr_id_to_sg_road_objects: Dict[int, List[Road]] = {}
    xodr_lane_to_sg_lane: Dict[xodrLane, Lane] = {}

    for road in road_network.get_roads():
        roads, old_to_new_lanes = road_to_sg(road, simplify_tolerance)
        xodr_id_to_sg_road_objects[road.id] = roads
        xodr_lane_to_sg_lane.update(old_to_new_lanes)

    for xodr_lane, sg_lane in xodr_lane_to_sg_lane.items():
        successor_lanes = xodr_lane.traffic_flow_successors
        for successor_lane in successor_lanes:
            try:
                successor_sg_lane = xodr_lane_to_sg_lane[successor_lane]
            except KeyError:
                raise KeyError(
                    f"Could not find successor lane {successor_lane} in "
                    + "OpenDRIVE to Scenario Gym dict; one of the successors of "
                    + f"{xodr_lane}."
                )
            add_connection((sg_lane, successor_sg_lane))

    return list(set().union(*xodr_id_to_sg_road_objects.values()))



import warnings
from typing import Dict, List, Tuple

import numpy as np
from numpy.linalg import norm
from shapely.geometry import LineString, Polygon

from scenario_gym.callback import StateCallback
from scenario_gym.entity import Entity
from scenario_gym.metrics.rss.rss_utils import (
    acceleration,
    ahead,
    coord_change,
    direction,
    inverse_direction,
)
from scenario_gym.state import State


class RSSParameters:
    """RSS parameters."""

    RESPONSE_TIME = 0.6  # SECONDS
    MIN_LONG_ACCEL = 1.2 * 9.81  # METRES PER SECONDS SQUARED
    MAX_LONG_ACCEL = 1.2 * 9.81  # METRES PER SECONDS SQUARED
    MIN_SAFE_CLEARANCE = 0.1  # METRES
    SHADOW_LENGTH = 100  # METRES
    VISIBLE_RADIUS = 50  # METRES
    LANE_ANGLE_VARIATION = 0.985  # COS(ANGLE)
    TIME_HORIZON = 3  # SECONDS


class RSSDistances(StateCallback):
    """
    Determine if the ego-entity distance has become unsafe.

    This is the per-timestep pre-computation for the first two rules of the
    RSS metric. Raw values are calculated and attached to the global state or
    entity attributes. Unsafe distances are flagged per entity in the global
    state, which the metric uses to return a boolean value per rule.
    """

    def _reset(self, state: State) -> None:
        """Reset callback and declares variables."""
        self.ego = state.scenario.entities[0]
        self.entities = state.scenario.entities
        # Initialise default callback parameters
        self.ego_params = {}
        self.entity_params = [{} for _ in self.entities[1:]]
        self.safe_distances = [[0.0, 0.0] for _ in self.entities[1:]]
        self.intersect = [["safe"] for _ in self.entities[1:]]
        self.entity_safe_ratios = {
            entity: [float("inf"), float("inf")] for entity in self.entities
        }

    def __call__(self, state: State) -> None:
        """
        Label each entity to specify if its distances to the ego are unsafe.

        Collates current timestep data for each entity, and determines the
        safety of the ego with respect to its longitudinal and lateral distance
        to each entity.

        --  Creates entity parameter dictionary from current pose and bounding box
        --  Calculates safe lateral and longitudinal distance to each entity
        --  Calculates safe distance ratios for rss-based colour-coding
        --  Determines, for each entity, if its position intersects the
            corresponding ego safe buffer, and if so, which direction should be
            flagged as unsafe
        """
        if state.t == 0.0:
            # Require at least two poses to calculate velocity
            return

        # Establish Ego parameters
        ego = self.ego
        entities = self.entities
        ego_heading = direction(state.poses[ego][3])
        ego_inverse_heading = inverse_direction(list(ego_heading))
        entity_params = []
        ego_position = state.poses[ego][0:2]
        # Create a dictionary for each entity of form:
        # {position, heading, velocity, acceleration, box_points, length, width}
        # With directional and positional parameters defined with respect to ego
        # frame.
        for entity in entities:
            entity_dictionary = self.get_entity_parameters(
                state,
                entity,
                ego_heading,
                ego_inverse_heading,
                ego_position,
                state.dt,
            )
            if entity_dictionary is None:
                warnings.warn(
                    "RSSDistances _step: Error in handling of entity {0} at time "
                    "{1}. Invalid pose: {2}. RSS metric skips entity at this "
                    "timestep.".format(entity, state.t, state.poses[entity])
                )
                continue
            else:
                entity_params.append(entity_dictionary)
        if not entity_params:
            warnings.warn(
                "Zero entity parameters generated at timestep: {0}".format(state.t)
            )
            return

        # Calculate safe distances between ego and all other entities
        ego_params = entity_params.pop(0)
        safe_distances = []
        for entity in entity_params:
            safe_long = abs(self.safe_longitudinal_distance(ego_params, entity))
            safe_lat = abs(self.safe_lateral_distance(ego_params, entity))
            safe_distances.append([safe_lat, safe_long])

        # Assign evaluated parameters to the state
        assert len(entity_params) == len(safe_distances)
        self.ego_params = ego_params
        self.entity_params = entity_params
        self.safe_distances = safe_distances

        # Check for each entity if there is an intersection of safe distance buffer
        # and assign safe distance ratios to entity as an attribute
        for i in range(len(entity_params)):
            self.entity_safe_ratios[entities[i + 1]] = self.safe_ratios(
                ego_params, entity_params[i], safe_distances[i]
            )
            self.intersect[i].append(
                self.unsafe_distance(
                    ego_params,
                    entity_params[i],
                    self.intersect[i],
                    safe_distances[i],
                )
            )
            if self.intersect[i] is None:
                # Entity trimmed poses results in IndexError, go to next entity
                warnings.warn(
                    "safe_longitudinal: IndexError in handling of entity number: "
                    "{0} at timestep: {1} seconds. Continue to next "
                    "entity.".format(i, state.t)
                )

    @staticmethod
    def safe_ratios(
        ego: Dict,
        haz: Dict,
        safe_distances: List[float],
    ) -> List[float]:
        """
        Attach safe_distance ratios to entity.

        safe ratio defined as actual_distance / safe_distance -> larger ratio safer
        """
        try:
            safe_lat = safe_distances[0] + 0.5 * abs(
                np.dot(
                    [haz["width"], haz["length"]], inverse_direction(haz["heading"])
                )
            )
            safe_long = safe_distances[1] + 0.5 * abs(
                np.dot([haz["width"], haz["length"]], haz["heading"])
            )
        except IndexError:
            # safe_distances not calculated
            warnings.warn(
                "RSSDistances safe_ratios: Safe distances not calculated. Default "
                "safe distances as ego dimensions"
            )
            safe_lat = 0.5 * ego["width"]
            safe_long = 0.5 * ego["length"]

        actual_lat = max(
            1e-6,
            abs(haz["position"][0])
            - 0.5 * ego["width"]
            - 0.5
            * abs(
                np.dot(
                    [haz["width"], haz["length"]], inverse_direction(haz["heading"])
                )
            ),
        )
        actual_long = max(
            1e-6,
            abs(haz["position"][1])
            - 0.5 * ego["length"]
            - 0.5 * abs(np.dot([haz["width"], haz["length"]], haz["heading"])),
        )
        return [abs(actual_lat / safe_lat), abs(actual_long / safe_long)]

    @staticmethod
    def unsafe_distance(
        ego: Dict, haz: Dict, intersect: List[str], safe_distances: List[float]
    ) -> str:
        """
        Determine if longitudinal or lateral distance is hazardous to ego.

        This method is called per timestep per entity, and returns a flag to be
        attached to the global state.

        --  Laterally unsafe if both longitudinal and lateral distances are unsafe,
            and the lateral distance became unsafe after the longitudinal distance.
        --  Longitudinally unsafe if both longitudinal and lateral distances are
            unsafe, and the longitudinal distance became unsafe after the lateral
            distance.
        """
        if "unsafe_lateral" in intersect or "unsafe_longitudinal" in intersect:
            # Already found
            intersect.append("found")
            return intersect

        buffer, lengths, widths = RSSDistances.generate_buffer(ego, safe_distances)

        assert (
            buffer.area > 0.0
        ), "safe_longitudinal: buffer constructed as a 'Z' rather than as a '[]'"

        hazard_area = Polygon(haz["box_points"])
        if hazard_area.intersects(buffer):
            # Find which direction became unsafe last: this is the unsafe direction
            for j in range(len(intersect), 0, -1):
                try:
                    if intersect[j - 1] == "lateral":
                        # longitudinal intersection last: longitudinally unsafe
                        return "unsafe_longitudinal"
                    elif intersect[j - 1] == "longitudinal":
                        # lateral intersection last: laterally unsafe
                        return "unsafe_lateral"
                except IndexError:
                    break
                # default, if no previous intersection found.
                if j == 1:
                    ego_dim = [ego["width"], ego["length"]]
                    if abs(
                        abs(haz["position"][0])
                        - abs(np.dot(haz["position"], ego_dim))
                    ) / safe_distances[0] > abs(
                        abs(
                            haz["position"][1]
                            - np.dot(
                                haz["position"],
                                inverse_direction(ego_dim),
                            )
                        )
                        / safe_distances[1]
                    ):
                        return "unsafe_longitudinal"
                    else:
                        return "unsafe_lateral"
        # This entity does not intersect buffer, so check if a dimension intersects
        return RSSDistances.write_intersections(lengths, widths, haz)

    @staticmethod
    def safe_longitudinal_distance(ego: Dict, haz: Dict) -> float:
        """Determine if the longitudinal distance is safe."""
        MAX_LONG_ACCEL = RSSParameters.MAX_LONG_ACCEL
        MIN_LONG_ACCEL = RSSParameters.MIN_LONG_ACCEL
        MIN_SAFE_CLEARANCE = RSSParameters.MIN_SAFE_CLEARANCE
        RESPONSE_TIME = RSSParameters.RESPONSE_TIME
        ego_direction = ego["heading"]
        hazard_direction = haz["heading"]
        ego_velocity = ego["velocity"]
        hazard_velocity = haz["velocity"]
        max_long_accel = abs(
            MAX_LONG_ACCEL * np.dot(ego_direction, hazard_direction)
        )
        if np.dot(ego_direction, hazard_direction) > 0:
            # Moving in the same direction (longitudinal component)
            if ahead(ego, haz):
                vf = norm(ego_velocity)
                vr = np.dot(hazard_velocity, ego_direction)
            else:
                vf = np.dot(hazard_velocity, ego_direction)
                vr = norm(ego_velocity)
            if vr == 0.0:
                # If rear car is stationary, safe
                return MIN_SAFE_CLEARANCE + 0.5 * ego["length"]
            d0 = RSSDistances.long_dist_same_direction(
                vf, vr, max_long_accel, RESPONSE_TIME, MIN_LONG_ACCEL
            )
        else:
            # Moving in the opposite direction (longitudinal component)
            v1 = abs(np.dot(ego_velocity, ego_direction))
            v2 = -abs(np.dot(hazard_velocity, ego_direction))
            # Makes no distinction between driver in correct or incorrect lane,
            # where RSS is specific
            if np.sign(haz["position"][1]) == np.sign(haz["velocity"][1]):
                return MIN_SAFE_CLEARANCE + 0.5 * ego["length"]
            d0 = RSSDistances.long_dist_opp_direction(
                v1, v2, max_long_accel, RESPONSE_TIME, MIN_LONG_ACCEL
            )
        return d0 + MIN_SAFE_CLEARANCE + 0.5 * ego["length"]

    @staticmethod
    def safe_lateral_distance(ego: Dict, haz: Dict) -> float:
        """Determine if the lateral distance is safe."""
        MAX_LONG_ACCEL = RSSParameters.MAX_LONG_ACCEL
        MIN_LONG_ACCEL = RSSParameters.MIN_LONG_ACCEL
        MIN_SAFE_CLEARANCE = RSSParameters.MIN_SAFE_CLEARANCE
        RESPONSE_TIME = RSSParameters.RESPONSE_TIME
        haz_position = np.array(haz["position"])
        # Resolve into velocity components para and perp to hazard's velocity
        # Velocity takes form [longitudinal, lateral]
        # Define lateral velocity as ego's velocity component perp to hazard
        v = haz["velocity"][0]  # component perpendicular to ego's heading
        max_lat_accel = MAX_LONG_ACCEL * abs(
            np.dot(inverse_direction(ego["heading"]), haz["heading"])
        )
        min_lat_accel = MIN_LONG_ACCEL * abs(
            np.dot(inverse_direction(ego["heading"]), haz["heading"])
        )
        if np.sign(-haz_position[0]) == np.sign(v):
            # lateral convergence
            v = abs(v)
            if v == 0.0:
                # Driving parallel, safe for sufficient constant distance.
                return MIN_SAFE_CLEARANCE + 0.5 * ego["width"]
            d0 = RSSDistances.lat_dist(
                v, max_lat_accel, min_lat_accel, RESPONSE_TIME
            )
        else:
            # lateral divergence: safe distance is the avg of car widths
            # plus min_safe_distance
            d0 = 0
        return d0 + MIN_SAFE_CLEARANCE + 0.5 * ego["width"]

    @staticmethod
    def write_intersections(
        buffer_lengths: List[LineString],
        buffer_widths: List[LineString],
        haz_dict: Dict,
    ) -> str:
        """
        Flag buffer intersection direction.

        For ego and another entity, check if the entity has a smaller distance in
        the longitudinal or lateral direction than the corresponding safe distance,
        and appends this to the recorded list.
        """
        haz_area = Polygon(haz_dict["box_points"])
        lat_inter = False
        long_inter = False

        if haz_area.intersects(buffer_lengths[0]) or haz_area.intersects(
            buffer_lengths[1]
        ):
            lat_inter = True
        if haz_area.intersects(buffer_widths[0]) or haz_area.intersects(
            buffer_widths[1]
        ):
            long_inter = True

        if lat_inter and long_inter:
            new_record = "both"
        elif lat_inter:
            new_record = "lateral"
        elif long_inter:
            new_record = "longitudinal"
        else:
            new_record = "safe"
        return new_record

    @staticmethod
    def get_entity_parameters(
        state: State,
        entity: Entity,
        ego_heading: List[float],
        ego_inverse_heading: List[float],
        ego_position: List[float],
        dt: float,
    ) -> Dict:
        """Calculate entity parameters and returns these as a dictionary."""
        entity_pose = state.poses[entity]
        entity_velocity = state.velocities[entity]
        if len(entity_pose) != 6:
            warnings.warn(
                "Entity pose should have six elements, [x, y, z, h, r, p]. "
                "Received {0} elements.".format(len(entity_pose))
            )
            return
        ego_position = np.array(ego_position)
        entity_heading = direction(entity_pose[3])
        entity_acceleration = acceleration(state.recorded_poses(entity), dt)
        # All vectors take form [lateral, longitudinal] / [x, y]
        entity_dictionary = {
            "position": coord_change(entity_pose[0:2], ego_heading, ego_position),
            "heading": [
                np.dot(entity_heading, ego_inverse_heading),
                np.dot(entity_heading, ego_heading),
            ],
            "velocity": [
                np.dot(entity_velocity[:2], ego_inverse_heading),
                np.dot(entity_velocity[:2], ego_heading),
            ],
            "accel": [
                np.dot(
                    entity_acceleration,
                    ego_inverse_heading,
                ),
                np.dot(entity_acceleration, ego_heading),
            ],
            "box_points": [
                coord_change(point, ego_heading, ego_position)
                for point in entity.get_bounding_box_points(entity_pose)
            ],
            "length": entity.catalog_entry.bounding_box.length,
            "width": entity.catalog_entry.bounding_box.width,
        }
        return entity_dictionary

    @staticmethod
    def generate_buffer(
        ego: Dict, safe_distances: List
    ) -> Tuple[Polygon, List[LineString]]:
        """
        Generate ego safe buffer corresponding to entity's safe distances.

        Generates a rectangular buffer around the entity with length and width
        corresponding to safe longitudinal and lateral distances with respect to
        the other entity. Returns the Polygon of this buffer along with a list of
        linestrings of its lengths and widths.
        """
        assert ego["position"] == [0.0, 0.0], ego["position"]
        try:
            safe_longitudinal_distance = safe_distances[1]
            safe_lateral_distance = safe_distances[0]
        except IndexError:
            # Safe distances not calculated
            warnings.warn(
                "RSSDistances generate_buffer: Safe distances not calculated: "
                "Buffer cannot be instantiated. Default safe distances as 3 "
                "metres lateral, 5 metres safe longitudinal"
            )
            safe_longitudinal_distance = 5
            safe_lateral_distance = 3

        buffer_vector = [
            np.array([0, safe_longitudinal_distance]),
            np.array([safe_lateral_distance, 0]),
        ]
        buffer = [
            np.array(buffer_vector[0] + buffer_vector[1]),
            np.array(buffer_vector[0] - buffer_vector[1]),
            np.array(-buffer_vector[0] - buffer_vector[1]),
            np.array(-buffer_vector[0] + buffer_vector[1]),
        ]
        widths = [
            LineString(
                [
                    [100 * buffer[0][0], buffer[0][1]],
                    [100 * buffer[1][0], buffer[1][1]],
                ]
            ),
            LineString(
                [
                    [100 * buffer[2][0], buffer[2][1]],
                    [100 * buffer[3][0], buffer[3][1]],
                ]
            ),
        ]
        lengths = [
            LineString(
                [
                    [buffer[0][0], 100 * buffer[0][1]],
                    [buffer[2][0], 100 * buffer[2][1]],
                ]
            ),
            LineString(
                [
                    [buffer[1][0], 100 * buffer[1][1]],
                    [buffer[3][0], 100 * buffer[3][1]],
                ]
            ),
        ]
        return Polygon(buffer), lengths, widths

    @staticmethod
    def long_dist_same_direction(
        vf: float,
        vr: float,
        max_long_accel: float,
        RESPONSE_TIME: float,
        MIN_LONG_ACCEL: float,
    ) -> float:
        """Return the minimum safe longitudinal distance for same directions."""
        return max(
            0,
            vr * RESPONSE_TIME
            + min(
                vf**2 / (2 * max_long_accel),
                0.5 * max_long_accel * RESPONSE_TIME**2,
            )
            + (vr + RESPONSE_TIME * max_long_accel) ** 2 / (2 * MIN_LONG_ACCEL)
            - vf**2 / (2 * max_long_accel),
        )

    @staticmethod
    def long_dist_opp_direction(
        v1: float,
        v2: float,
        max_long_accel: float,
        RESPONSE_TIME: float,
        MIN_LONG_ACCEL: float,
    ) -> float:
        """Return the minimum safe longitudinal distance for opposing directions."""
        return max(
            0,
            (
                (2 * v1 + RESPONSE_TIME * max_long_accel) * RESPONSE_TIME / 2
                + (v1 + RESPONSE_TIME * max_long_accel) ** 2 / (2 * MIN_LONG_ACCEL)
                + (2 * abs(v2) + RESPONSE_TIME * max_long_accel) * RESPONSE_TIME / 2
                + (abs(v2) + RESPONSE_TIME * max_long_accel) ** 2
                / (2 * MIN_LONG_ACCEL)
            ),
        )

    @staticmethod
    def lat_dist(
        v: float, max_lat_accel: float, min_lat_accel: float, RESPONSE_TIME: float
    ):
        """Return the minimum safe lateral distance between the entity and ego."""
        return max(
            0,
            0.5 * RESPONSE_TIME * (2 * v + RESPONSE_TIME * max_lat_accel)
            + (v + RESPONSE_TIME * max_lat_accel) ** 2 / (2 * min_lat_accel)
            - 0.5 * RESPONSE_TIME**2 * max_lat_accel
            - (RESPONSE_TIME * max_lat_accel) ** 2 / (2 * min_lat_accel),
        )



from abc import ABC, abstractclassmethod
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from lxml.etree import Element

from scenario_gym.utils import ArgsKwargs


@dataclass(frozen=True)
class Catalog:
    """A catalog."""

    catalog_name: str
    rel_path: str


class CatalogObject(ABC):
    """
    Base class for objects loaded from catalogs.

    Subclasses should implement the `load_data_from_xml` class method wwhich takes
    the specific xml element that contains the data and returns the arguements
    and keyword arguments for the class constructor. It should not return the
    class itself since this way the methods can make use of loading methods from
    parent classes.

    The attribute xosc_namex can be set to the element names that the object
    represents. For example, if `xosc_names = ["Vehicle"]` then any elements with
    the tag `Vehicle` will be loaded by this entry. If not set then the class name
    will be used.
    """

    xosc_names: Optional[List[str]] = None

    @classmethod
    def from_xml(cls, catalog: Catalog, element: Element):
        """Create the class from an xml element."""
        args, kwargs = cls.load_data_from_xml(catalog, element)
        return cls(*args, **kwargs)

    @abstractclassmethod
    def load_data_from_xml(cls, catalog: Catalog, element: Element) -> ArgsKwargs:
        """Load the object from an xml element."""
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create the object from a dictionary.

        Must be implemented to allow json serialization.
        """
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        """
        Write the object to a dictionary.

        Must be implemented to allow json serialization.
        """
        raise NotImplementedError


@dataclass
class BoundingBox(CatalogObject):
    """A bounding box defined by its length, width and center."""

    width: float
    length: float
    center_x: float
    center_y: float

    @classmethod
    def load_data_from_xml(cls, catalog: Catalog, element: Element) -> ArgsKwargs:
        """Load the bounding box data form an xml element."""
        if element.tag != "BoundingBox":
            raise TypeError(f"Expected BoundingBox element not {element.tag}.")
        bb_center = element.find("Center")
        bb_dimensions = element.find("Dimensions")
        return (
            float(bb_dimensions.attrib["width"]),
            float(bb_dimensions.attrib["length"]),
            float(bb_center.attrib["x"]),
            float(bb_center.attrib["y"]),
        ), {}

    @classmethod
    def from_dict(cls, data: Dict[str, float]):
        """Load the bounding box from a dictionary."""
        return cls(
            data["width"],
            data["length"],
            data["center_x"],
            data["center_y"],
        )

    def to_dict(self) -> Dict[str, float]:
        """Write the bounding box to a jsonable dictionary."""
        return {
            "width": self.width,
            "length": self.length,
            "center_x": self.center_x,
            "center_y": self.center_y,
        }


@dataclass
class CatalogEntry(CatalogObject):
    """
    A single catalog entry. Holds catalog information and a bounding box.

    Parameters
    ----------
    catalog : Catalog
        The catalog from which the entry is loaded.

    catalog_entry : str
        The name of the specific entry.

    catalog_category : Optional[str]
        The category of the entry.

    catalog_type : str
        The catalog type e.g Vehicle or Pedestrian.

    properties : Dict[str, Union[float, str]]
        Any properties associated with the element.

    files: List[str]
        A list of filepaths for external files.

    """

    catalog: Catalog
    catalog_entry: str
    catalog_category: Optional[str]
    catalog_type: str
    bounding_box: BoundingBox
    properties: Dict[str, Union[float, str]]
    files: List[str]

    @classmethod
    def load_data_from_xml(cls, catalog: Catalog, element: Element) -> ArgsKwargs:
        """Load the catalog entry from an xml element."""
        entry_name = element.attrib["name"]
        cname = element.tag.lower() + "Category"
        category = element.attrib[cname] if cname in element.attrib else None
        bb = element.find("BoundingBox")
        bb = BoundingBox.from_xml(catalog, bb)
        properties, files = cls.load_properties_from_xml(element)
        return (
            catalog,
            entry_name,
            category,
            element.tag,
            bb,
            properties,
            files,
        ), {}

    @staticmethod
    def load_properties_from_xml(
        element: Element,
    ) -> Tuple[Dict[str, Union[str, float]], List[str]]:
        """
        Load properties from the xml element.

        These can be either `Property` or `File` elements. `Property` elements are
        given with `name` and `value` attributes (`name` must be unique for the
        entry) and are returned as a dict of values indexed by `name`. The value
        will be converted to a float if it can otherwise the string will be
        returned. `File` elements must have a `filepath` attribute which will
        be parsed. Multiple files can be stored with one entity.
        """
        files = []
        properties = {}
        prop = element.find("Properties")
        if prop is not None:
            for child in prop.findall("Property"):
                try:
                    v = child.attrib["value"]
                    with suppress(ValueError):
                        v = float(v)
                    properties[child.attrib["name"]] = v
                except KeyError as e:
                    raise RuntimeError(
                        "Property could not be loaded without `value` key."
                    ) from e
            for file in prop.findall("File"):
                files.append(file.attrib["filepath"])
        return properties, files

    @property
    def catalog_name(self) -> str:
        """Get the name of the catalog file."""
        return self.catalog.catalog_name

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the catalog entry from a dictionary."""
        return cls(
            Catalog(data["catalog"]["catalog_name"], data["catalog"]["rel_path"]),
            data["catalog_entry"],
            data["catalog_category"],
            data["catalog_type"],
            BoundingBox.from_dict(data["bounding_box"]),
            data.get("properties", {}),
            data.get("files", []),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the catalog entry to a dictionary."""
        return {
            "catalog": {
                "catalog_name": self.catalog.catalog_name,
                "rel_path": self.catalog.rel_path,
            },
            "catalog_entry": self.catalog_entry,
            "catalog_category": self.catalog_category,
            "catalog_type": self.catalog_type,
            "bounding_box": self.bounding_box.to_dict(),
            "properties": self.properties,
            "files": self.files,
        }



from __future__ import annotations

import json
import warnings
from contextlib import suppress
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Tuple, Type

import matplotlib.pyplot as plt
import numpy as np

from scenario_gym.entity import Entity
from scenario_gym.entity.pedestrian import Pedestrian
from scenario_gym.entity.vehicle import Vehicle
from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario.actions import ScenarioAction
from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import cached_property


class Scenario:
    """
    The scenario_gym representation of a scenario.

    A scenario consists of a set of entities and a road network. The entities have
    trajectories and catalog entries and may have additional entity specific
    properties. The scenario also may have a list of actions which can specify
    events that occur. E.g. a traffic light changing color.
    """

    def __init__(
        self,
        entities: List[Entity],
        name: Optional[str] = None,
        path: Optional[str] = None,
        road_network: Optional[RoadNetwork] = None,
        actions: Optional[List[ScenarioAction]] = None,
        properties: Optional[Dict[Any, Any]] = None,
    ):
        self._entities = entities
        self._ref_to_entity: Dict[str, Entity] = {e.ref: e for e in entities}

        self.name = name
        self.path = path
        self.road_network = road_network
        self.actions = actions if actions is not None else []
        self.properties = properties if properties is not None else {}

        self._vehicles: Optional[List[Entity]] = None
        self._pedestrians: Optional[List[Entity]] = None
        self._catalog_locations: Optional[Dict[str, str]] = None

    @property
    def entities(self) -> List[Entity]:
        """Get the entities in the scenario."""
        return self._entities

    @property
    def vehicles(self) -> List[Entity]:
        """Get the entities that have vehicle catalogs."""
        if self._vehicles is None:
            self._vehicles = [e for e in self.entities if isinstance(e, Vehicle)]
        return self._vehicles

    @property
    def catalog_locations(self) -> Dict[str, str]:
        """Get the locations of each catalog."""
        if self._catalog_locations is None:
            catalogs = set((e.catalog_entry.catalog for e in self.entities))
            self._catalog_locations = {
                catalog.catalog_name: catalog.rel_path for catalog in catalogs
            }
        return self._catalog_locations

    @property
    def pedestrians(self) -> List[Entity]:
        """Get the entities that have pedestrian catalogs."""
        if self._pedestrians is None:
            self._pedestrians = [
                e for e in self.entities if isinstance(e, Pedestrian)
            ]
        return self._pedestrians

    @property
    def trajectories(self) -> Dict[str, Trajectory]:
        """Return a dictionary mapping entity references to the trajectory."""
        return {e.ref: e.trajectory for e in self.entities}

    @cached_property
    def length(self) -> float:
        """Return the length of the scenario in seconds."""
        return max([t.max_t for t in self.trajectories.values()])

    def entity_by_name(self, e_ref: str) -> Optional[Entity]:
        """Return an entity given a unique reference."""
        with suppress(KeyError):
            return self._ref_to_entity[e_ref]

    def __copy__(self) -> Scenario:
        """Create a copy of a scenario without copying the road network."""
        return self.__class__(
            name=f"Copy of {self.name}" if self.name is not None else None,
            road_network=self.road_network,
            actions=deepcopy(self.actions),
            entities=[e.copy() for e in self.entities],
            properties=self.properties,
        )

    def copy(self) -> Scenario:
        """Create a copy of the scenario."""
        return copy(self)

    def add_entity(self, e: Entity, inplace: bool = False) -> Scenario:
        """Create a new scenario with the entity added."""
        if e.ref in self._ref_to_entity:
            i = 0
            while True:
                new_ref = f"{e.ref}_{i}"
                if new_ref not in self._ref_to_entity:
                    break
                i += 1
            old_ref = e.ref
            e.ref = new_ref
            warnings.warn(
                f"An entity with ref {old_ref} exists. Adding with ref {new_ref}."
            )
        scenario = self.copy() if not inplace else self
        scenario._entities.append(e)
        scenario._ref_to_entity[e.ref] = e
        scenario._vehicles = None
        scenario._pedestrians = None
        return scenario

    def remove_entity(self, e: Entity, inplace: bool = False) -> Scenario:
        """Create a new scenario with the entity added."""
        scenario = self.copy() if not inplace else self
        scenario._entities.remove(e)
        scenario._ref_to_entity.pop(e.ref)
        scenario._vehicles = None
        scenario._pedestrians = None
        return scenario

    def make_ego(self, e: Entity, inplace: bool = False) -> Scenario:
        """Set e to the ego entity."""
        try:
            idx = self._entities.index(e)
        except ValueError:
            idx = None
        scenario = self.copy() if not inplace else self

        if idx is not None:
            e = scenario._entities.pop(idx)
        else:
            scenario._ref_to_entitiy[e.ref] = e
        scenario._entities.insert(0, e)
        scenario.vehicles = None
        scenario.pedestrians = None
        return scenario

    def add_action(self, action: ScenarioAction, inplace: bool = False) -> Scenario:
        """Add an action to the scenario."""
        scenario = self.copy() if not inplace else self
        scenario.actions.append(action)
        return scenario

    def translate(self, x: np.ndarray, inplace: bool = False) -> Scenario:
        """Return a new scenario with all entities translated."""
        scenario = self.copy() if not inplace else self
        for e in scenario.entities:
            e.trajectory = e.trajectory.translate(x)
        return scenario

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        e_classes: Tuple[Type[Entity]] = (Vehicle, Pedestrian, Entity),
        a_classes: Tuple[Type[ScenarioAction]] = (ScenarioAction,),
    ):
        """Load the scenario from a dictionary."""
        entities = []
        for e_data in data["entities"]:
            for Ent in e_classes:
                if Ent.__name__ == e_data.get("entity_class", Entity):
                    break
            entities.append(Ent.from_dict(e_data))

        road_network = data.get("road_network")
        if road_network is not None:
            road_network = RoadNetwork.create_from_json(road_network["path"])

        actions = []
        for a_data in data.get("actions", ()):
            for Act in a_classes:
                if Act.__name__ == a_data.get("action_class", ScenarioAction):
                    break
            actions.append(Act.from_dict(a_data))

        return cls(
            entities,
            name=data.get("name"),
            path=data.get("path"),
            road_network=road_network,
            actions=actions,
            properties=data.get("properties", {}),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the scenario to a dictionary."""
        return {
            "entities": [e.to_dict() for e in self.entities],
            "name": self.name,
            "path": self.path,
            "road_network": {
                "name": self.road_network.name,
                "path": self.road_network.path,
            }
            if self.road_network is not None
            else None,
            "actions": [act.to_dict() for act in self.actions],
            "properties": self.properties,
        }

    @classmethod
    def from_json(
        cls,
        path: str,
        e_classes: Tuple[Type[Entity]] = (Vehicle, Pedestrian, Entity),
        a_classes: Tuple[Type[ScenarioAction]] = (ScenarioAction,),
    ):
        """Load the scenario from a json file."""
        with open(path, "r") as f:
            data = json.load(f)
        return cls.from_dict(data, e_classes=e_classes, a_classes=a_classes)

    def to_json(self, path) -> None:
        """Write the scenario to a json file."""
        with open(path, "w") as f:
            json.dump(self.to_dict(), f)

    def describe(self) -> None:
        """Generate a text overview of the scenario."""
        rn = self.road_network.path.split("/")[-1].split(".")[0]
        name = (
            self.name.replace(".xosc", "")
            if self.name is not None
            else self.path.split("/")[-1].split(".")[0]
        )
        title = f"Scenario: {name}"
        under_header = "=" * len(title)

        entity_header = "Entity".ljust(10) + "Type".ljust(10) + "Cateogry".ljust(10)
        entities = ""
        for e in self.entities:
            entities += (
                f"{e.ref}".ljust(10)
                + f"{e.type}".ljust(10)
                + f"{e.catalog_entry.catalog_category}".ljust(10)
                + "\n"
            )

        print(
            f"""
{title}
{under_header}
Filepath: {self.path}
Road network: {rn}
Number of entities: {len(self.entities)}
Total duration: {self.length:.4}s

Entities
--------
{entity_header}
{entities}
"""
        )

    def plot(self, figsize: Tuple[int, int] = (10, 10), show: bool = True) -> None:
        """
        Visualise the scenario.

        Parameters
        ----------
        figsize : Tuple[int, int]
            The figure size.

        show : bool
            If set to False will not call `plt.show` so the figure can be modified
            or saved.

        """
        name = (
            self.name.replace(".xosc", "")
            if self.name
            else self.scenario_path.split("/")[-1].split(".")[0]
        )
        plt.figure(figsize=figsize)
        for geom in self.road_network.driveable_surface.geoms:
            plt.fill(*geom.exterior.xy, c="gray", alpha=0.25)
            for i in geom.interiors:
                plt.fill(*i.xy, c="white")
        for r in self.road_network.roads:
            plt.plot(*r.center.xy, c="white")
        for i, e in enumerate(self.entities):
            c = "r" if i == 0 else "b"
            plt.plot(*e.trajectory.data[:, [1, 2]].T, c=c, label=e.ref)
            plt.plot(*e.trajectory.data[0, [1, 2]].T, c=c, marker="o")
        data = np.vstack([e.trajectory.data[:, [1, 2]] for e in self.entities])
        b_min, b_max = data.min(0), data.max(0)
        plt.axis("equal")
        plt.xlim(b_min[0] - 10.0, b_max[0] + 10.0)
        plt.ylim(b_min[1] - 10.0, b_max[1] + 10.0)
        plt.legend()
        plt.title(name)
        if show:
            plt.show()



import math
from enum import Enum
from typing import List, Optional, Tuple

import numpy as np
from shapely.geometry import Polygon

from scenario_gym.entity import Entity
from scenario_gym.metrics.base import Metric
from scenario_gym.state import State


def angle_between(x: float, a_low: float, a_high: float) -> bool:
    """Return True if angle x is between a_low and a_high."""
    x = x % (math.pi * 2)
    a_low = a_low % (math.pi * 2)
    a_high = a_high % (math.pi * 2)
    return (
        ((a_low < x) or (x <= a_high))
        if (a_low >= a_high)
        else (a_low <= x < a_high)
    )


class CollisionTypes(Enum):
    """Enumerates possible collision types."""

    other = 0
    t_bone = 1
    head_on = 2
    rear_end = 3
    side_swipe = 4
    non_vehicle = 5


class CollisionPoints(Enum):
    """Enumerates possible collision points around a bounding box."""

    front = 0
    front_corner = 1
    side = 2
    back = 3
    back_corner = 4


class CollisionMetric(Metric):
    """
    Detects and classifies collisions with the ego.

    Records all collisions between entities and the ego. If the
    hazard entity is a vehicle then the collision is classified
    into t_bone, head_on, rear_end or side_swipe. If not then it
    is recorded as other.
    """

    name = "collisions"

    def __init__(self, c_tol: float = 0.4, name: Optional[str] = None):
        self.ego: Optional[Entity] = None
        self.collisions: List[Tuple[float, str, CollisionTypes]] = []
        self.c_tol = c_tol
        super().__init__(name=name)

    def _reset(self, state: State) -> None:
        """Reset the ego and recorded collisions."""
        ego = state.scenario.entity_by_name("ego")
        if ego is None:
            ego = state.scenario.entities[0]
        self.ego = ego
        self.collisions: List[Tuple[float, str, CollisionTypes]] = []
        self.last_timestep: List[Entity] = []

    def _step(self, state: State) -> None:
        """Update recorded collisions."""
        for e_other in state.collisions()[self.ego]:
            if e_other not in self.last_timestep:
                self.collisions.append(self.record_collision(state, e_other))
        self.last_timestep = state.collisions()[self.ego].copy()

    def get_state(self) -> List[Tuple[float, str, str]]:
        """Return the recorded collisions."""
        return [(t, ref, c.name) for t, ref, c in self.collisions]

    def record_collision(
        self, state: State, hazard: Entity
    ) -> Tuple[float, str, CollisionTypes]:
        """Classify the collision and record it."""
        if hazard.catalog_entry.catalog_type != "Vehicle":
            return (state.t, hazard.ref, CollisionTypes.non_vehicle)

        ego_box = self.ego.get_bounding_box_geom(state.poses[self.ego])
        hazard_box = hazard.get_bounding_box_geom(state.poses[hazard])

        collision_point = np.array(
            ego_box.intersection(hazard_box).centroid.xy
        ).squeeze()
        collision_angle = (hazard.pose[3] - self.ego.pose[3]) % (math.pi * 2)

        ego_angle = (
            np.arctan2(*np.flip(collision_point - self.ego.pose[:2]))
            - self.ego.pose[3]
        ) % (math.pi * 2)
        hazard_angle = (
            np.arctan2(*np.flip(collision_point - hazard.pose[:2])) - hazard.pose[3]
        ) % (math.pi * 2)

        ego_point = self.get_collision_point(ego_box, ego_angle, self.ego.pose[3])
        hazard_point = self.get_collision_point(
            hazard_box, hazard_angle, hazard.pose[3]
        )

        ego_front = ego_point in (
            CollisionPoints.front,
            CollisionPoints.front_corner,
        )
        ego_back = ego_point in (CollisionPoints.back, CollisionPoints.back_corner)
        hazard_front = hazard_point in (
            CollisionPoints.front,
            CollisionPoints.front_corner,
        )
        hazard_back = hazard_point in (
            CollisionPoints.back,
            CollisionPoints.back_corner,
        )

        if ego_front and hazard_front:
            if angle_between(
                collision_angle,
                math.pi / 4,
                3 * math.pi / 4,
            ) or angle_between(
                collision_angle,
                5 * math.pi / 4,
                7 * math.pi / 4,
            ):
                ctype = CollisionTypes.t_bone
            elif angle_between(
                collision_angle,
                7 * math.pi / 4,
                math.pi / 4,
            ):
                ctype = CollisionTypes.side_swipe
            else:
                ctype = CollisionTypes.head_on
        elif (ego_front or ego_back) and (hazard_front or hazard_back):
            if angle_between(
                collision_angle,
                math.pi / 4,
                3 * math.pi / 4,
            ) or angle_between(
                collision_angle,
                5 * math.pi / 4,
                7 * math.pi / 4,
            ):
                ctype = CollisionTypes.t_bone
            else:
                ctype = CollisionTypes.rear_end
        elif any([ego_front, ego_back, hazard_front, hazard_back]):
            if angle_between(
                collision_angle,
                math.pi / 4,
                3 * math.pi / 4,
            ) or angle_between(
                collision_angle,
                5 * math.pi / 4,
                7 * math.pi / 4,
            ):
                ctype = CollisionTypes.t_bone
            else:
                ctype = CollisionTypes.side_swipe
        else:
            ctype = CollisionTypes.side_swipe

        return state.t, hazard.ref, ctype

    def get_collision_point(
        self,
        box: Polygon,
        angle: float,
        heading: float,
    ) -> CollisionPoints:
        """Classify the collision point of the angle given the bounding box."""
        c_tol = self.c_tol
        corners = (
            np.arctan2(  # corners are BL, FL, FR, BR
                *np.flip(
                    np.array(box.exterior.coords[:-1]).T
                    - np.array(box.centroid.xy),
                    axis=0,
                ),
            )
            - heading
        )
        if (angle_between(angle, corners[1] - c_tol, corners[1] + c_tol)) or (
            angle_between(angle, corners[2] - c_tol, corners[2] + c_tol)
        ):
            return CollisionPoints.front_corner
        elif (angle_between(angle, corners[0] - c_tol, corners[0] + c_tol)) or (
            angle_between(angle, corners[3] - c_tol, corners[3] + c_tol)
        ):
            return CollisionPoints.back_corner
        elif angle_between(angle, corners[0] + c_tol, corners[3] - c_tol):
            return CollisionPoints.back
        elif angle_between(angle, corners[2] - c_tol, corners[1] + c_tol):
            return CollisionPoints.front
        return CollisionPoints.side


class CollisionPointMetric(Metric):
    """
    Finds the co-ordinate, and relative agles of collision.

    Records the position of any collisions that
    occur between entities and the ego. Collects
    the (x, y) coordinate and relative angle [rad]
    between the entities at the collision point.
    """

    name = "collision_points"

    def __init__(self, name: Optional[str] = None):
        self.ego: Optional[Entity] = None
        self.collisions: List[Tuple[str, np.ndarray, float]] = []
        super().__init__(name=name)

    def _reset(self, state: State) -> None:
        """Reset the ego and recorded collisions."""
        ego = state.scenario.entity_by_name("ego")
        if ego is None:
            ego = state.scenario.entities[0]
        self.ego = ego
        self.collisions: List[Tuple[str, np.ndarray, float]] = []
        self.last_timestep: List[Entity] = []

    def _step(self, state: State) -> None:
        """Update recorded collision angle and position."""
        for e_other in state.collisions()[self.ego]:
            if e_other not in self.last_timestep:
                self.collisions.append(
                    self.record_collision_position(state, e_other)
                )
        self.last_timestep = state.collisions()[self.ego].copy()

    def get_state(self) -> List[Tuple[str, np.ndarray, float]]:
        """Return the entity reference, coordinates and angle of collisions."""
        return self.collisions

    def record_collision_position(
        self, state: State, hazard: Entity
    ) -> Tuple[str, np.ndarray, float]:
        """Calculate the coordinate and relative angle of entities at collision."""
        ego_box = self.ego.get_bounding_box_geom(state.poses[self.ego])
        hazard_box = hazard.get_bounding_box_geom(state.poses[hazard])

        collision_point = np.array(
            ego_box.intersection(hazard_box).centroid.xy
        ).squeeze()
        collision_angle = (hazard.pose[3] - self.ego.pose[3]) % (math.pi * 2)
        return hazard.ref, collision_point, collision_angle



from dataclasses import dataclass
from typing import Any, Dict, Optional

from lxml.etree import Element

from scenario_gym.catalog_entry import (
    ArgsKwargs,
    BoundingBox,
    Catalog,
    CatalogEntry,
)
from scenario_gym.entity.base import Entity
from scenario_gym.trajectory import Trajectory


@dataclass
class PedestrianCatalogEntry(CatalogEntry):
    """Catalog entry for a pedestrian."""

    mass: Optional[float]

    xosc_names = ["Pedestrian"]

    @classmethod
    def load_data_from_xml(cls, catalog_name: str, element: Element) -> ArgsKwargs:
        """Load the vehicle from an xml element."""
        base_args, _ = super().load_data_from_xml(catalog_name, element)
        mass = element.attrib.get("mass")
        if mass is not None:
            mass = float(mass)
        return base_args + (mass,), {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the pedestrian from a dictionary."""
        return cls(
            Catalog(data["catalog"]["catalog_name"], data["catalog"]["rel_path"]),
            data["catalog_entry"],
            data["catalog_category"],
            data["catalog_type"],
            BoundingBox.from_dict(data["bounding_box"]),
            data.get("properties", {}),
            data.get("files", []),
            data.get("mass"),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the pedestrian to a dictionary."""
        data = super().to_dict()
        data["mass"] = self.mass
        return data


class Pedestrian(Entity):
    """Entity class for pedestrians."""

    def __init__(
        self,
        catalog_entry: PedestrianCatalogEntry,
        trajectory: Optional[Trajectory] = None,
        ref: Optional[str] = None,
    ):
        super().__init__(catalog_entry, trajectory=trajectory, ref=ref)



from scenario_gym.metrics.base import Metric, cache_mean, cache_metric
from scenario_gym.metrics.collision import CollisionMetric
from scenario_gym.metrics.rss.rss import RSS, RSSDistances
from scenario_gym.metrics.trajectory import (
    EgoAvgSpeed,
    EgoDistanceTravelled,
    EgoMaxSpeed,
)
